#include <u.h>
#include <libc.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>

#define BUCKSIZE 1048576
#define MAXQ 1024
#define SMBUF 256

typedef struct Hub	Hub;
typedef struct Msgq	Msgq;

struct Hub{
	char name[SMBUF];			/* name */
	char bucket[BUCKSIZE];		/* data buffer */
	char *inbuckp;				/* location to store next message */
	int buckfull;				/* amount of data stored in bucket */
	Req *qreqs[MAXQ];			/* pointers to queued Reqs */
	int rstatus[MAXQ];			/* status of requests */
	int qrnum;				/* index of Reqs waiting to be filled */
	int qans;					/* number of Reqs answered */
};

struct Msgq{
	ulong myfid;
	char *nxt;
	int bufuse;
};

enum {
	UP = 1,
	DOWN = 2,
	WAIT = 3,
	DONE = 4,
};

static char Ebad[] = "something bad happened";
static char Enomem[] = "no memory";

void msgsend(Hub *h);
void zerohub(Hub *h);
void fsread(Req *r);
void fswrite(Req *r);
void fscreate(Req *r);
void fsopen(Req *r);
void fsdestroyfile(File *f);
void usage(void);

Srv fs = {
	.open=	fsopen,
	.read=	fsread,
	.write=	fswrite,
	.create=	fscreate,
};

void
msgsend(Hub *h)
{
	Msgq *mq;
	u32int count;

	for(int i = h->qans; i <= h->qrnum; i++){
		if(h->rstatus[i] != WAIT){
			if(i == h->qans){
				h->qans++;
			}
			continue;
		}
		mq = h->qreqs[i]->fid->aux;
		if(mq->nxt != h->inbuckp){
			count = h->qreqs[i]->ifcall.count;
			if(mq->bufuse + count >= BUCKSIZE - 8192){
				mq->nxt = h->bucket;
				mq->bufuse = 0;
			}
			if((mq->bufuse + count > h->buckfull) && (mq->bufuse < h->buckfull)){
				count = h->buckfull - mq->bufuse;
			}
			memmove(h->qreqs[i]->ofcall.data, mq->nxt, count);
			h->qreqs[i]->ofcall.count = count;
			mq->nxt += count;
			mq->bufuse += count;
			h->rstatus[i] = DONE;
			if(i == h->qans){
				h->qans++;
			}
			respond(h->qreqs[i], nil);
		}
	}
}

void
fsread(Req *r)
{
	Hub *h;

	h = r->fid->file->aux;
	h->qrnum++;
	if(h->qrnum >= MAXQ - 2){
		msgsend(h);
		h->qrnum = 1;
		h->qans = 0;
	}
	h->rstatus[h->qrnum] = WAIT;
	h->qreqs[h->qrnum] = r;
	msgsend(h);
}

void
fswrite(Req *r)
{
	u32int count;
	Hub *h;

	h = r->fid->file->aux;
	count = r->ifcall.count;
	if((h->buckfull + count) >= BUCKSIZE - 8192){
		h->inbuckp = h->bucket;
		h->buckfull = 0;
	}

	memmove(h->inbuckp, r->ifcall.data, count);
	h->inbuckp += count;
	h->buckfull += count;
	r->fid->file->length = h->buckfull;
	r->ofcall.count = count;
	msgsend(h);
	respond(r, nil);
}

void
fscreate(Req *r)
{
	Hub *h;
	File *f;

	if(f = createfile(r->fid->file, r->ifcall.name, r->fid->uid, r->ifcall.perm, nil)){
		h = emalloc9p(sizeof(Hub));
		zerohub(h);
		strcat(h->name, r->ifcall.name);
		f->aux = h;
		r->fid->file = f;
		r->ofcall.qid = f->qid;
		respond(r, nil);
		return;
	}
	respond(r, Ebad);
}

void
fsopen(Req *r)
{
	Hub *h;
	Msgq *q;

	h = r->fid->file->aux;
	q = (Msgq*)emalloc9p(sizeof(Msgq));
	memset(q, 0, sizeof(Msgq));
	q->myfid = r->fid->fid;
	q->nxt = h->bucket;
	q->bufuse = 0;
	r->fid->aux = q;
	if(h && (r->ifcall.mode&OTRUNC)){
		h->inbuckp = h->bucket;
		h->buckfull = 0;
		r->fid->file->length = 0;
	}
	respond(r, nil);
}

void
fsdestroyfile(File *f)
{
	Hub *h;

	h = f->aux;
	if(h){
		free(h);
	}
}

void
zerohub(Hub *h)
{
	memset(h, 0, sizeof(Hub));
	h->inbuckp = h->bucket;
	h->qrnum = 0;
	h->qans = 0;
}

void
usage(void)
{
	fprint(2, "usage: hubfs [-D] [-s srvname] [-m mtpt]\n");
	exits("usage");
}

void
main(int argc, char **argv)
{
	char *addr = nil;
	char *srvname = nil;
	char *mtpt = nil;
	Qid q;
	fs.tree = alloctree(nil, nil, DMDIR|0777, fsdestroyfile);
	q = fs.tree->root->qid;

	ARGBEGIN{
	case 'D':
		chatty9p++;
		break;
	case 'a':
		addr = EARGF(usage());
		break;
	case 's':
		srvname = EARGF(usage());
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	default:
		usage();
	}ARGEND;

	if(argc)
		usage();
	if(chatty9p)
		fprint(2, "hubsrv.nopipe %d srvname %s mtpt %s\n", fs.nopipe, srvname, mtpt);
	if(addr == nil && srvname == nil && mtpt == nil)
		sysfatal("must specify -a, -s, or -m option");
	if(addr)
		listensrv(&fs, addr);
	if(srvname || mtpt)
		postmountsrv(&fs, srvname, mtpt, MREPL|MCREATE);
	exits(0);
}
